#include "aspps.h"

void resetMetaAtomCounts(Node *atom);
int evaluateClause(Node *clause);
int getCurrentHornCount(Node *atom);
void pushAtomStack(Node *atom, int value);
void pushClauseStack(Node *satisfiedClause, int value);
void setAtomWeight(Node *atom);


void parseVariableText(char *input)
{
	int		i,len;
	int		index;
	char	buffer[1000000];

	len=strlen(input)-1;
	i=0;
	while(i<len&&!isdigit((int)input[i])&&input[i]!='-') i++;
	index=atoi(&input[i]);
	if(input[i]=='-') i++;
	while(i<len&&isdigit((int)input[i])) i++;
	input[len]='\0';
	while(i<len&&isspace((int)input[i])) i++;
	
	if(index>0)
		atomList[index].atomText= strdup(&input[i]);
	else{
		if(index==0) 
			sprintf(buffer,"%s\n%s",atomList[0].atomText,&input[i]); 
		else
			sprintf(buffer,"%s\n!%s",atomList[0].atomText,&input[i]); 
		free(atomList[0].atomText);
		atomList[0].atomText=strdup(buffer);
	}
}

int parseInteger(char *input, int *begin)
{
	int i,j;
	char line[180];

	i=*begin;
	while(input[i]!=EOLN&&isspace((int)input[i]))i++;
	if(input[i]==EOLN || input[i]=='}')
		return(-1);
	j=0;
	while(isdigit((int)input[i])) line[j++]=input[i++];
	line[j]=EOLN;
	(*begin)=i;
	return(atoi(line));
}

Node **getList(char *input,int *index,int *count)
{
	int		i,j,size;
	Node	**tmp=NULL;

	size=0;
	i=(*index);
	while((parseInteger(input,&i)>=0)) size++;
	if(size){
		tmp=(Node **)calloc(sizeof(Node *),size);
		(*count)=size;
		i=(*index);
		for(j=0;j<size;j++){
			tmp[j]=&atomList[parseInteger(input,&i)];
		}
	}
	(*index)=i;
	return(tmp);
}

Node *parseAtom(char *input,int *index)
{
	int	   count=0,pos,neg;
	char	buffer[180];
	Node **list;
	

	(*index)++;
	pos=parseInteger(input,index);
	neg=parseInteger(input,index);
	list=getList(input,index,&count);
	if(Debug){
		if(list)
			fprintf(errorFile,"pos=%i neg=%i list[0]=%i count=%i\n",
					pos,neg,list[0]->index,count);
		else
			fprintf(errorFile,"pos=%i neg=%i list=NULL count=%i\n",pos,neg,count);
	}
	if(Maxlit<count) Maxlit=count;
	numberOfMetaAtoms++;
	sprintf(buffer,"Meta%i",numberOfMetaAtoms);
	atomList[numberOfMetaAtoms].atomText= strdup(buffer);
	atomList[numberOfMetaAtoms].nType=META;
	atomList[numberOfMetaAtoms].requiredPositive=pos;
	if(count-neg<0)
		atomList[numberOfMetaAtoms].requiredNegative=0;
	else atomList[numberOfMetaAtoms].requiredNegative=count-neg;
	atomList[numberOfMetaAtoms].metaList=list;
	atomList[numberOfMetaAtoms].metaCount=count;
	return(&atomList[numberOfMetaAtoms]);
}

int parseCNFClause(char *input,int index)
{
	int		i=0,count =0;
	int endOfClause=0;
	int state = 0;
	
	while(i<strlen(input)&&!endOfClause){
		switch(state){
			case 0:
				if(isdigit((int)input[i])){
					if(input[i]=='0'){
						endOfClause = 1;
					}
					else{
						clauseList[index].headCount++;
						state = 1;
					}
				}
				else if(input[i] == '-'){
						clauseList[index].bodyCount++;
						state = 1;
				}
				break;
			case 1:
				if(isspace((int)input[i]))
					state = 0;
				break;
			default:
				break;
		}
		i++;
	}
	if(clauseList[index].headCount+clauseList[index].bodyCount==0) return(False);
	clauseList[index].nType=CLAUSE;
	requiredSatisfiedClauses++;
	clauseList[index].headList=(Node  **)calloc(sizeof(Node *),clauseList[index].headCount);
	clauseList[index].bodyList=(Node  **)calloc(sizeof(Node *),clauseList[index].bodyCount);
	clauseList[index].headCount=clauseList[index].bodyCount=0;

	i=0;
	while(i<strlen(input)){
		if(input[i]=='-'){
			i++;
			clauseList[index].bodyList[clauseList[index].bodyCount++]=&atomList[parseInteger(input,&i)];
		}
		else if(isdigit((int)input[i])&& input[i]!='0'){
				clauseList[index].headList[clauseList[index].headCount++]=&atomList[parseInteger(input,&i)];
		}
		i++;
	}

	count=clauseList[index].headCount+clauseList[index].bodyCount;
	if(Maxlit<count)
		Maxlit=count;
	else if(Minlit>count)
		Minlit=count;
	return(1);
}

int parseClause(char *input,int index)
{
	int		i=0,count=0,hd=0;
	
	
	clauseList[index].nType=CLAUSE;
	while(i<strlen(input)){
		if(input[i]=='{'){
			if(hd)
				clauseList[index].headCount++;
			else clauseList[index].bodyCount++;
			while(input[i]!='}') i++;
		}
		else if(input[i]==',')
			hd=1;
		else if(input[i]==':'){
			hd=1;
			clauseList[index].nType=HORNCLAUSE;
		}
		else if(isdigit((int)input[i])){
			if(hd)clauseList[index].headCount++;
			else clauseList[index].bodyCount++;
			while(isdigit((int)input[i]))i++;
		}
		i++;
	}
	if(clauseList[index].headCount+clauseList[index].bodyCount==0) return(False);
	clauseList[index].headList=(Node  **)calloc(sizeof(Node *),clauseList[index].headCount);
	clauseList[index].bodyList=(Node  **)calloc(sizeof(Node *),clauseList[index].bodyCount);
	clauseList[index].headCount=clauseList[index].bodyCount=i=hd=0;
	while(i<strlen(input)){
		if(input[i]==','||input[i]==':')
			hd=1;
		else if(input[i]=='{'){
			if(hd)
				clauseList[index].headList[clauseList[index].headCount++]=parseAtom(input,&i);
			else
				clauseList[index].bodyList[clauseList[index].bodyCount++]=parseAtom(input,&i);
		}
		else if(isdigit((int)input[i])){
			if(hd)
				clauseList[index].headList[clauseList[index].headCount++]=&atomList[parseInteger(input,&i)];
			else
				clauseList[index].bodyList[clauseList[index].bodyCount++]=&atomList[parseInteger(input,&i)];
		}

		i++;
	}

	if(clauseList[index].nType==CLAUSE)
		requiredSatisfiedClauses++;
	count=clauseList[index].headCount+clauseList[index].bodyCount;
	if(Maxlit<count)
		Maxlit=count;
	else if(Minlit>count)
		Minlit=count;
	return(1);
}
		
void setCnt(void)
{
	int i,increase=1;;

	for(i=1;i<=numberOfAtoms;i++)
		if(atomList[i].assignedValue==Unknown)
			increase*=2;
	Cnt+=increase;
}

void printAnswer(void)
{
	int	i,j,k;
	char	buffer[180];

	if(Cnt==0)
		fprintf(stdout,
				"\nAtoms determined during grounding and don't care atoms. Don't care atoms are preceded by !: %s\n"
				,atomList[0].atomText);
	for(j=0;j<ShowCnt;j++)
		printf(" %s\n",ShowAtoms[j]);
	fprintf(stdout,"\nAnswer-set %i:\n",Cnt+1);
	for(i=1;i<=numberOfAtoms;i++)
		if(atomList[i].assignedValue!=False){
			if(!ShowCnt){
				if(atomList[i].assignedValue==Unknown){
					printf(" not %s or",atomList[i].atomText);
				}
				printf(" %s\n",atomList[i].atomText);
			}
			else{
				j=0;
				while(j<strlen(atomList[i].atomText)&&(isdigit(atomList[i].atomText[j])||isspace(atomList[i].atomText[j])))
					j++;
				k=0;
				buffer[k]='\0';
				while(j<strlen(atomList[i].atomText)&&atomList[i].atomText[j]!='(')
					buffer[k++]=atomList[i].atomText[j++];
				buffer[k]='\0';
				j=0;
				while(j<ShowCnt){
					if(!strcmp(buffer,ShowAtoms[j])){
						if(atomList[i].assignedValue==Unknown)
							printf(" not %s or",atomList[i].atomText);
						printf(" %s\n",atomList[i].atomText);
					}
					j++;
				}
			}
		}
	fprintf(stdout,"\n");
}

int verifyInputType(char *line, int *i)
{
	int j;
	char theoryType[180];

	while(isspace(line[*i]))
		(*i)++;
	if(isdigit(line[*i]))
		return(1);
	j=0;
	while(isalpha(line[*i]))
		theoryType[j++]=line[(*i)++];
	theoryType[j]='\0';
	if(!strcmp(theoryType,"aspps"))
		return(1);
	else if (!strcmp(theoryType,"cnf"))
		return(2);
	return(0);
	
}

													   
int getProblem(int *sol)
{
	int		index,i;
	char	line[1800];
	int metaAtomCount=0;
	int formatType;


	index=0;
	requiredSatisfiedClauses=0;
	while (fgets(line,1799,fptr)&&line[0]!='p');
	i=1; 
	if(line[0]=='p' && (formatType = verifyInputType(line,&i))){
		numberOfAtoms = numberOfMetaAtoms =parseInteger(line,&i);
		if(Debug){
			fprintf(errorFile, "numberOfAtoms = % i\n", numberOfAtoms);
		}
		i++;
		if(formatType == 1){
			metaAtomCount =  parseInteger(line,&i);
			i++;
		}
		allAtoms= numberOfAtoms + metaAtomCount;
		if(Debug){
			fprintf(errorFile, "allAtoms = % i\n", allAtoms);
		}
		numberOfClauses=parseInteger(line,&i);
		if(Debug){
			fprintf(errorFile, "numberOfClauses = % i\n", numberOfClauses);
		}
	}
	else{
		printf("error: input format not aspps\n");
		return(0);
	}
	clauseStackSize = numberOfClauses * 2;
	atomStackSize = allAtoms * 2;
	requiredAtomsStackSize = allAtoms * 2;
	branchingStackSize = allAtoms + 1;
	atomList=(Node *)calloc(sizeof(Node),(allAtoms+1));
	atomStack=(Node **)calloc(sizeof(Node *),(atomStackSize));
	requiredAtomStack=(Node **)calloc(sizeof(Node *),(requiredAtomsStackSize));
	branchingStack=(int *)calloc(sizeof(int ),(branchingStackSize));
	atomList[0].atomText=strdup("");
	clauseList=(Node *)calloc(sizeof(Node),numberOfClauses);
	clauseStack=(Node **)calloc(sizeof(Node *),(clauseStackSize));
	literalList=(int *)calloc(sizeof(int),(allAtoms+1));
	while (fgets(line,1799,fptr)){
		switch(line[0]){
			case 'p': break;
			case 'c': if(formatType ==1) parseVariableText(line); break;
			default: 
				if(formatType == 1){
					if(!parseClause(line,index++)) { *sol = 0; return(1);} 
				}
				else if(!parseCNFClause(line,index++)){ *sol = 0;  return(1);}
				break;
		}
	}
	if(fptr != stdin)
		fclose(fptr);
	if(numberOfClauses!=index||numberOfMetaAtoms!=allAtoms){
		printf("Input Error:  numberOfClauses=%i index=%i -- numberOfMetaAtoms=%i allAtoms=%i\n",
				numberOfClauses,index,numberOfMetaAtoms,allAtoms);
		return(0);
	}
	if(numberOfClauses==0){ *sol = -1; return(1); }
	propagateQueue=(Node **)calloc(sizeof(Node *),(allAtoms+1));
	propagateQueueValue=(int *)calloc(sizeof(int),(allAtoms+1));
	hornPathList=(int *)calloc(sizeof(int ),(allAtoms+1));
	Weight = (unsigned long *)calloc(sizeof(unsigned long),(Maxlit+1));
	if(Debug){
		fprintf(errorFile,
			"Maxlit = %i Minlit = %i requiredSatisfiedClauses= %i numberOfClauses = %i allAtoms = % i\n",
			Maxlit,Minlit,requiredSatisfiedClauses,numberOfClauses,allAtoms);
	}
	*sol = 1;
	return(1);
}

void initializeCurrent(void)
{
	int i;

	if(Debug)
		fprintf(errorFile, "initializeCurrent\n");
	currentSatisfiedClauses = 0;
	numberOfRequiredHornAtoms = 0;
	atomTop=0;
	clauseTop=0;
	branchTop=0;
	requiredAtomTop=0;
	atomsAssigned=0;
	for(i=0;i<numberOfClauses;i++){
		clauseList[i].currentBodyCount=clauseList[i].bodyCount;
		clauseList[i].currentHeadCount=clauseList[i].headCount;
		clauseList[i].currentCount=clauseList[i].headCount+clauseList[i].bodyCount;
		clauseList[i].assignedValue=False;
	}
	for(i=1;i<=allAtoms;i++){
		atomList[i].currentPositive=atomList[i].requiredPositive;
		atomList[i].currentNegative=atomList[i].requiredNegative;
		atomList[i].currentCount=atomList[i].metaCount;
		atomList[i].assignedValue=Unknown;
	}
	for(i=1;i<numberOfClauses;i++)
		clauseList[i].weight = Weight[clauseList[i].currentCount];
	for(i=1;i<numberOfAtoms;i++)
		setAtomWeight(&atomList[i]);
}	


void setAtomWeight(Node *atom)
{
	int j;

	atom->weight=0;
	for(j=0;j<atom->headCount;j++)
			if( ((atom->weight + atom->headList[j]->weight)> atom->weight))
				atom->weight += atom->headList[j]->weight;
	for(j=0;j<atom->bodyCount;j++)
			if( ((atom->weight + atom->bodyList[j]->weight)> atom->weight))
				atom->weight += atom->bodyList[j]->weight;
	for(j=0;j<atom->metaCount;j++)
			if( ((atom->weight + Weight[atom->metaList[j]->currentCount]) > atom->weight))
				atom->weight += Weight[atom->metaList[j]->currentCount];
}

void setWeights(void)
{
	int i,j;
	int overflow ;
	double Mult,wgt;
	double increment=3;

	Mult = 13;
	Weight[0]=1;
	Weight[1]=1;
	do{
		j=1;
		wgt=Mult;
		overflow=0;
		Weight[Maxlit]=1;
		i=Maxlit-1;
		while(i>1 && !overflow){
			wgt = j * Mult;
			 Weight[i]  = (unsigned long)rint( wgt);
			if( Weight[i] < Weight[i+1])
				overflow=1;
			i--;
			j*=increment;
		}
		if(Mult > 1)
			Mult -= 0.5;
		else{
			Mult = 13;
			increment-=1;
		}
	}while(overflow);
	
	if(Debug)
		for(i=0;i<=Maxlit;i++)
			fprintf(stderr, "WeightList[%i]=%lu \n",i, Weight[i]);
}

int setPropagatedAtom(Node *atom, int value)
{
	if(atom->assignedValue==Unknown){  
		if(!atom->isRequired || atom->requiredValue == value){
			pushAtomStack(atom,value);
			if(atom->nType==HORNATOM && atom->isRequired == True)
				numberOfRequiredHornAtoms++;
		}
		else return False;
	}
	return True;
}

void pushAtomStack(Node *atom, int value)
{
	if(atomTop>=atomStackSize){
		atomStackSize*=2;
		atomStack=realloc(atomStack,atomStackSize*sizeof(Node  *));
	}
	if(atom == NULL)	
		atomStack[atomTop++]=NULL;
	else{
		atom->assignedValue = value;
		atomStack[atomTop++]=atom;
		atomsAssigned++;
	}

}

void pushRequiredAtomStack(Node *atom,int requiredValue)
{
	if(requiredAtomTop>=requiredAtomsStackSize){
		requiredAtomsStackSize*=2;
		requiredAtomStack=realloc(requiredAtomStack,requiredAtomsStackSize*sizeof(Node  *));
	}
	if(atom==NULL)
		requiredAtomStack[requiredAtomTop++]=NULL;
	else{
		if(atom->isRequired) return;
		if(atom->nType==HORNATOM) 
			numberOfRequiredHornAtoms++;
		atom->isRequired=True;
		atom->requiredValue=requiredValue;
		requiredAtomStack[requiredAtomTop++]=atom;
	}
}

void pushStacks(void)
{
	pushAtomStack(NULL,0);
	pushRequiredAtomStack(NULL,0);
	pushClauseStack(NULL,0);
}

void pushBranchValues( int count)
{
	int i;
	if(branchTop + count >= branchingStackSize){
		branchingStackSize*=2;
		branchingStack=realloc(branchingStack,sizeof(int)*branchingStackSize);
	}
	for(i=0;i<count;i++){
		branchingStack[branchTop++]=literalList[i];
		pushStacks();
		if(Debug)
			fprintf(errorFile, " lit = %i ", literalList[i]);
	}
}

void pushClauseStack(Node *satisfiedClause, int value)
{
	if(clauseTop >= clauseStackSize){
		clauseStackSize*=2;
		clauseStack=(Node **)realloc(clauseStack,sizeof(Node *)*clauseStackSize);
	}
	if(satisfiedClause == NULL)
		clauseStack[clauseTop++]=NULL;
	else{
		if(satisfiedClause->assignedValue == False)
			clauseStack[clauseTop++]=satisfiedClause;
		if((value == True) && (satisfiedClause->nType==CLAUSE))
			currentSatisfiedClauses++;
		satisfiedClause->assignedValue = value;
	}
}

void resetClauseCount(Node *clause)
{
	int i;
	
	clause->currentHeadCount=clause->headCount;
	clause->currentBodyCount=clause->bodyCount;
	clause->currentCount=clause->bodyCount+clause->headCount;
	for(i=0;i<clause->headCount;i++)
		if(clause->headList[i]->assignedValue!=Unknown){
			clause->currentHeadCount--;
			clause->currentCount--;
		}

	for(i=0;i<clause->bodyCount;i++)
		if(clause->bodyList[i]->assignedValue!=Unknown){
			clause->currentBodyCount--;
			clause->currentCount--;
		}
	clause->weight = Weight[clause->currentCount];
}

void resetMetaAtomWeight(Node *atom)
{
	int i;

	atom->weight=0;
	for(i=0;i<atom->metaCount;i++){
		if( (atom->metaList[i]->assignedValue==Unknown) &&
				(atom->weight + atom->metaList[i]->weight > atom->weight))
			atom->weight += atom->metaList[i]->weight;
	}
}

void resetMetaAtomCounts(Node *atom)
{
	int i;

	atom->currentCount=atom->metaCount;
	atom->currentPositive=atom->requiredPositive;
	atom->currentNegative=atom->requiredNegative;
	for(i=0;i<atom->metaCount;i++){
		if(atom->metaList[i]->assignedValue==True){
			atom->currentPositive--;
			atom->currentCount--;
		}
		else if(atom->metaList[i]->assignedValue==False){
			atom->currentNegative--;
			atom->currentCount--;
		}
	}
}

int popBranchStack(void)
{
	if (branchTop > 0){
		branchTop--;
		return(branchingStack[branchTop]);
	}
	return(0);
}


int popBacktrackingStacks(void)
{

	if(branchTop>0){
		while(atomTop>0&&atomStack[--atomTop]){
			if(atomStack[atomTop]->assignedValue!=Unknown){
				atomStack[atomTop]->assignedValue=Unknown;
				if(atomStack[atomTop]->nType==HORNATOM&&
					atomStack[atomTop]->isRequired==True&&
					atomStack[atomTop]->requiredValue==True)
						numberOfRequiredHornAtoms++;
				atomsAssigned--;
			}
		}
		while(requiredAtomTop>0&&requiredAtomStack[--requiredAtomTop]){
				requiredAtomStack[requiredAtomTop]->requiredValue=Unknown;
				requiredAtomStack[requiredAtomTop]->isRequired=False;
				if(requiredAtomStack[requiredAtomTop]->nType==HORNATOM)
					numberOfRequiredHornAtoms--;
		}
		while(clauseTop>0&&clauseStack[--clauseTop]){
			if(clauseStack[clauseTop]->assignedValue==True&&
				clauseStack[clauseTop]->nType==CLAUSE)
					currentSatisfiedClauses--;
			resetClauseCount(clauseStack[clauseTop]);
			clauseStack[clauseTop]->assignedValue=False;
		}
		return(1);
	}
	return(0);

}


void setReference(void)
{

	int i,j;
	Node *aptr;

	for(j=0;j<numberOfClauses;j++){
		for(i=0;i<clauseList[j].bodyCount;i++)
			clauseList[j].bodyList[i]->bodyCount++;
		for(i=0;i<clauseList[j].headCount;i++){
			clauseList[j].headList[i]->headCount++;
			if(clauseList[j].nType==HORNCLAUSE)
				clauseList[j].headList[i]->nType=HORNATOM;
		}
	}
	for(i=numberOfAtoms+1;i<=allAtoms;i++)
			for(j=0;j<atomList[i].metaCount;j++)
				atomList[i].metaList[j]->metaCount++;
	for(i=1;i<=allAtoms;i++){
		atomList[i].bodyList=(Node **)calloc(sizeof(Node *),(atomList[i].bodyCount));
		atomList[i].headList=(Node **)calloc(sizeof(Node *),(atomList[i].headCount));
		if(atomList[i].nType!=META){
			atomList[i].metaList=(Node **)calloc(sizeof(Node *),(atomList[i].metaCount));
			atomList[i].metaCount=0;
		}
		atomList[i].bodyCount=0;
		atomList[i].headCount=0;
	}
	for(j=0;j<numberOfClauses;j++){
		for(i=0;i<clauseList[j].bodyCount;i++)
			clauseList[j].bodyList[i]->bodyList[clauseList[j].bodyList[i]->bodyCount++]= &clauseList[j];
		for(i=0;i<clauseList[j].headCount;i++)
			clauseList[j].headList[i]->headList[clauseList[j].headList[i]->headCount++]= &clauseList[j];
		clauseList[j].index = j;
	}
	for(i=1;i<=allAtoms;i++){
		if(atomList[i].nType==META)
			for(j=0;j<atomList[i].metaCount;j++){
				aptr=atomList[i].metaList[j];
				aptr->metaList[aptr->metaCount++]=&atomList[i];
			}
		atomList[i].index=i;
	}
}

void printTheory(void)
{
	int	i,j;

	fprintf(errorFile," \n");
	for(i=0;i<numberOfClauses;i++){
		fprintf(errorFile,
				"Rule %i  %i: body count %i head count %i -- bodylist:",
				i,clauseList[i].nType,clauseList[i].bodyCount,clauseList[i].headCount);
		for(j=0;j<clauseList[i].bodyCount;j++)
			fprintf(errorFile," %i ",clauseList[i].bodyList[j]->index);
		fprintf(errorFile," -- headlist:");
		for(j=0;j<clauseList[i].headCount;j++)
			fprintf(errorFile," %i ",clauseList[i].headList[j]->index);
		fprintf(errorFile," \n");
	}

	fprintf(errorFile," \n");
	for(i=1;i<=allAtoms;i++){
		fprintf(errorFile,"Atom %i %s value = %i -- ",
				i,atomList[i].atomText, atomList[i].assignedValue);
			fprintf(errorFile,"Atomlist:");
		if(atomList[i].nType==META)
			fprintf(errorFile,"MetaList:");
		for(j=0;j<atomList[i].metaCount;j++)
			fprintf(errorFile," %i ",atomList[i].metaList[j]->index);
		fprintf(errorFile," \n");
	}
}

int computeMetaValue(Node *atom)
{
	if(atom->currentNegative<=0&&atom->currentPositive<=0)
		return(True);
	else if( (atom->currentCount<atom->currentPositive)||
			(atom->currentCount<atom->currentNegative)||
			(atom->currentCount<atom->currentPositive+atom->currentNegative))
		return(False);
	return(Unknown);
}
int determineSetValue(Node *atom,int value)
{
	if(value==True){
		if(atom->currentNegative<=0&&atom->currentPositive==atom->currentCount)
			return(True);
		else if(atom->currentPositive<=0&&atom->currentNegative==atom->currentCount)
			return(False);
	}
	else if(atom->currentCount==1){
		if(atom->currentNegative==1)
			return(True);
		else if(atom->currentPositive==1) 
			return(False);
	}
	return(Unknown);
}
int propagateMetaList(Node **metaList,int metaCount, int setValue)
{
	int i;
	for(i=0;i<metaCount;i++){
		if(metaList[i]->assignedValue==Unknown){
			if(propagateQueueValue[metaList[i]->index]==Unknown){
				propagateQueue[propagateCount++]=metaList[i];
				propagateQueueValue[metaList[i]->index]=setValue;
			}
			else if(propagateQueueValue[metaList[i]->index]!=setValue)
				return(False);
		}
	}
	return(True);
}

int setMeta(Node *atom, int value)
{
	int setValue=-1;
	int computedMetaValue=-1;
	 
	if(atom->isRequired&&atom->requiredValue!=value)
		return(0);
	if(Debug)
		fprintf(errorFile, "setMeta = %i cnt = %i pos =%i neg= %i val=%i\n",
				atom->index,atom->currentCount,atom->currentPositive,
				atom->currentNegative,value);
	pushRequiredAtomStack(atom,value);
	resetMetaAtomCounts(atom);
	computedMetaValue=computeMetaValue(atom);
	if(value==computedMetaValue) 
		return(True);
	else if(computedMetaValue!=Unknown)
		return(False);
	setValue=determineSetValue(atom,value);
	if(setValue==Unknown)
		return(setValue);
	return(propagateMetaList(atom->metaList,atom->metaCount,setValue));
} 

int setHorn(Node *ltl,int horn,int val)
{
	 
	if(ltl->isRequired==True&&ltl->requiredValue!=val)
		return(False);
	pushRequiredAtomStack(ltl,val);
	if(horn==0) return(Unknown);
	return(True);
} 

int  propagateAtoms(Node **list,int cnt,int horn, int val)
{
	int i,returnValue=True;

	for(i=0;i<cnt;i++){
		if(list[i]->assignedValue==Unknown){
			if(propagateQueueValue[list[i]->index]==Unknown){
				returnValue=True;
				if(list[i]->nType==HORNATOM)
					returnValue=setHorn(list[i],horn,val);
				else if(list[i]->nType==META)
					returnValue=setMeta(list[i],val);
				if(returnValue==True){
					propagateQueue[propagateCount++]=list[i];
					propagateQueueValue[list[i]->index]=val;
				}
				else if(returnValue==False) 
					return(False);
			}
			else if(propagateQueueValue[list[i]->index]!=val)
				return(False);
		}
	}
	return(returnValue);
}


int closure(Node *clause,int level)
{
	int i,j,hornCount;
	Node *atom;

	if(Debug)
		fprintf(errorFile, 
		"closure level %i rule = %i atom = %i \n"
		,level, clause->index,clause->headList[0]->index);
	
	for(i=0;i<clause->bodyCount;i++){
		atom=clause->bodyList[i];
		if(atom->assignedValue==Unknown){
			if(propagateQueueValue[atom->index]==Unknown){
				if(atom->nType==HORNATOM){
					if(Debug)
						fprintf(errorFile, "Pathlist = %i atom = %i \n"
						,hornPathList[atom->index], atom->index);
					if(hornPathList[atom->index])return(0);
					pushRequiredAtomStack(atom,True);
					hornCount=getCurrentHornCount(atom);
					if(hornCount==1){
						hornPathList[atom->index]=1;
						for(j=0;j<atom->headCount;j++){
							if(atom->headList[j]->nType==HORNCLAUSE&&
								atom->headList[j]->assignedValue!=True){
								if(closure(atom->headList[j],level+1)==0) return(0);
							}
						}
					}
					else if(hornCount==0) return(0);
				}
				else{
					propagateQueue[propagateCount++]=atom;
					propagateQueueValue[atom->index]=True;
				}
			}
			else if(propagateQueueValue[atom->index]==False)
				return(False);
		}
	}
	return(1);
}

int getCurrentHornCount(Node *atom)
{
	int i, hornCount=0;

	for(i=0;i<atom->headCount;i++)
		if(atom->headList[i]->nType==HORNCLAUSE&&
				atom->headList[i]->assignedValue!=True)
			hornCount++;
	return(hornCount);
}

int checkHorn(Node *atom)
{
	int i,hornCount=0;
	if(Debug)
		fprintf(errorFile, 
		"checkHorn atom %i ",atom->index);
	if(propagateQueueValue[atom->index]!=Unknown) return(True);
	hornCount=getCurrentHornCount(atom);

	if(hornCount>1)
		return(True);
	else if(hornCount == 0){
		if(atom->isRequired==True){
			if(atom->requiredValue==True)
				return(False);
		}
		else 
			pushRequiredAtomStack(atom,False);
		propagateQueue[propagateCount++]=atom;
		propagateQueueValue[atom->index]=False;
	}
	else if(hornCount==1&&
		atom->isRequired==True&&atom->requiredValue==True){
		memset(hornPathList,0,sizeof(int)*(allAtoms+1));
		hornPathList[atom->index]=1;
		for(i=0;i<atom->headCount;i++)
			if(atom->headList[i]->nType==HORNCLAUSE&&
				atom->headList[i]->assignedValue!=True)
				if(closure(atom->headList[i],0)==0) 
					return(False);
	}
	return(True);
}

int evaluateMetaList(Node *atom)
{
	int computedMetaValue = Unknown,setValue=Unknown;

	if(atom->assignedValue==Unknown&&propagateQueueValue[atom->index]==Unknown){
		resetMetaAtomCounts(atom);
		computedMetaValue=computeMetaValue(atom);
		if(atom->isRequired==False){
			if(computedMetaValue==Unknown)
				return(Unknown);
			else{ 
				propagateQueue[propagateCount++]=atom;
				propagateQueueValue[atom->index]=computedMetaValue;
			}
		}
		else{ 
			if(computedMetaValue!=Unknown){
					if(atom->requiredValue!=computedMetaValue)
						return(False);
					else{
						propagateQueue[propagateCount++]=atom;
						propagateQueueValue[atom->index]=atom->requiredValue;
					}
			}
			else{
				setValue=determineSetValue(atom,atom->requiredValue);
				if(setValue==Unknown)
					return(setValue);
				if(!propagateMetaList(atom->metaList,atom->metaCount,setValue))
					return(False);
				propagateQueue[propagateCount++]=atom;
				propagateQueueValue[atom->index]=atom->requiredValue;
			}
		}
	}
	return(True);
}

int reductMetaAtoms(Node *atom)
{
	int i;

	for(i=0;i<atom->metaCount;i++)
		if(evaluateMetaList(atom->metaList[i])==False)
			return(False);
	return(True);
}


int reductHeadAtoms(Node *atom, int value)
{
	int i;
	Node *clause;

	for(i=0;i<atom->headCount;i++){
		clause=atom->headList[i];
		if(clause->assignedValue!=True){
			if(value){
				pushClauseStack(clause,True);
				}
			else{
				clause->currentHeadCount--;
				pushClauseStack(clause,Modified);
				if(!evaluateClause(clause))
					return(False);
			}
		}
	}
	return(True);
}

int reductBodyAtoms(Node *atom, int value)
{
	int i;
	Node *clause;

	for(i=0;i<atom->bodyCount;i++){
		clause=atom->bodyList[i];
		if(clause->assignedValue!=True){
			if(value){
				clause->currentBodyCount--;
				pushClauseStack(clause,Modified);
				if(!evaluateClause(clause))
					return(False);
			}
			else{
				pushClauseStack(clause,True);
				if(clause->nType==HORNCLAUSE&&!checkHorn(clause->headList[0]))
					return(False);
			}
		}
	}
	return(True);
}

int propagateClauses(Node *atom,int value)
{

	if(!setPropagatedAtom(atom, value))
		return(False);
	
	if(atom->nType==ATOM)
		if(reductMetaAtoms(atom)==False)
				return(False);

	if(reductHeadAtoms(atom,value)==False)
		return(False);
	
	if(reductBodyAtoms(atom,value)==False)
		return(False);
	return(True);
}

int checkHornClosure(void)
{
	int i,j;
	int found=False;

	i=0;
	while(i<numberOfClauses){
		if(clauseList[i].assignedValue!=True && clauseList[i].nType==HORNCLAUSE){
			found=1;
			for(j=0;j<clauseList[i].bodyCount;j++)
				if(clauseList[i].bodyList[j]->nType==HORNATOM &&
					clauseList[i].bodyList[j]->assignedValue != True) 
						found=0;
		}
		if(found) return(True);
		i++;	
	}

	return(False);
}

int propagate(void)
{
	int i=0;
	int value,returnValue=True;

	while(i<propagateCount){
		value=propagateQueueValue[propagateQueue[i]->index];
		if((returnValue=propagateClauses(propagateQueue[i],value))==False)
			return(returnValue);
		i++;
	}
	if(Debug)
		fprintf(errorFile, "Propagate cnt = %i\n",propagateCount);
	if(currentSatisfiedClauses>=requiredSatisfiedClauses)
		return(-1);
	if(numberOfRequiredHornAtoms > 0 && (checkHornClosure() == False))
		return(False);
	return(returnValue);
}

int setupPropagation(void)
{
	int literal;
	propagateCount=0;
	memset(propagateQueueValue,Unknown,sizeof(int)*(allAtoms+1));
	if((literal=popBranchStack())!=0){
		propagateQueue[propagateCount]=&atomList[abs(literal)];
		if(literal<0)
			propagateQueueValue[-literal]=False;
		else
			propagateQueueValue[literal]=True;
		propagateCount++;
		return(propagate());
	}
	return(False);

}

int isBranchMetaAtom(Node *atom)
{
		if((atom->currentPositive==1&&
			(1+atom->currentNegative)==atom->currentCount)
			||(atom->currentNegative==1&&
			(1+atom->currentPositive)==atom->currentCount))
		return(1);
	return(0);
}
/*
int bestMetaAtom(int count)
{
	int i,j;
	unsigned int totalCount,maxCount=0;
	int metaIndex=0;
	Node *metaAtom=NULL, *atom=NULL;

	for(i=0;i<count;i++){
		totalCount=0;
		metaAtom=&atomList[literalList[i]];
		for(j=0;j<metaAtom->metaCount;j++){
			atom=metaAtom->metaList[j];
			if(atom->assignedValue==Unknown)
				totalCount+= atom->metaCount+atom->headCount + atom->bodyCount;
		}
		if(totalCount > maxCount){
			maxCount=totalCount;
			metaIndex=literalList[i];
		}
		else if((Random && totalCount == maxCount) && 
				(rand()>RAND_MAX/2)){
			metaIndex=literalList[i];
		}
	}
	return(metaIndex);
}
*/

int getMetaAtom(void)
{
	int j;
	int minMetaLength=Maxlit;
	int metaLength=0;
	int minOriginalLength=Maxlit;
	int metaOriginalLength=0;
	int metaAtom=0;
	double metaWgt = 0;
	double maxMetaWgt = 0;

	for(j=numberOfAtoms+1;j<=allAtoms;j++)
		if(atomList[j].assignedValue==Unknown && atomList[j].isRequired==True
				&&atomList[j].requiredValue==True){ 
			resetMetaAtomCounts(&atomList[j]);
			if(isBranchMetaAtom(&atomList[j])){
				metaLength=atomList[j].currentCount;
				if(metaLength <= minMetaLength){
					resetMetaAtomWeight(&atomList[j]);
					metaWgt=atomList[j].weight;
					metaOriginalLength = atomList[j].metaCount;
					if((metaLength < minMetaLength) || (maxMetaWgt < metaWgt) || 
							(metaOriginalLength < minOriginalLength)){
						minMetaLength = metaLength;
						maxMetaWgt = metaWgt;
						minOriginalLength = metaOriginalLength ;
						metaAtom=j;
					}
					else if((metaLength == minMetaLength) && (maxMetaWgt == metaWgt) && 
							(minOriginalLength == metaOriginalLength) && Random && (rand()>RAND_MAX/2))
						metaAtom=j;
				}
			}
		}
	return(metaAtom);
}

void sortLiteralList(int count)
{
	int i,j;
	int maxWeight=0;
	int maxIndex=0;
	int tmp;
	Node *atom;

	for(i=count-1;i>0;i--){
		maxIndex=0;
		atom = &atomList[literalList[0]];
		maxWeight= atom->weight;
		for(j=1;j<=i;j++){
			atom = &atomList[literalList[j]];
			if(maxWeight < atom->weight){
				maxWeight = atom->weight;
				maxIndex=j;
			}
		}
		if(i != maxIndex){
			tmp=literalList[i];
			literalList[i] = literalList[maxIndex];
			literalList[maxIndex]=tmp;
		}
	}
}

int setLiteralList(Node *atom)
{
	int j, literalListCount=0;
	

	if(Debug)
		fprintf(errorFile,"setLiteralList atom=%i\n", atom->index);


	for(j=0;j<atom->metaCount;j++)
		if(atom->metaList[j]->assignedValue==Unknown){
			literalList[literalListCount]=atom->metaList[j]->index;
			literalListCount++;
		}
	sortLiteralList(literalListCount);
	if(atom->currentPositive > 1)
		for(j=0;j<literalListCount;j++)
			literalList[j]= -literalList[j];
	pushBranchValues(literalListCount);
	return(literalListCount);
}


int getLiteral(void)
{
	int i, literal = 0;
	unsigned int max=0;
	
	for(i=1;i<=numberOfAtoms;i++){
		if(atomList[i].assignedValue==Unknown&&atomList[i].nType==ATOM){ 
			if(atomList[i].weight > max){
				max = atomList[i].weight;
				literal = i;
			}
			else if(Random && (atomList[i].weight == max)&&(rand()>RAND_MAX/2)){
					literal = i;
				}
		}
	}
	return(literal);
}


int getBranchLiterals(void)
{
	int literal=0;
	int metaAtom=0;
	int i,j;

	i=clauseTop;
	while(i>0&&clauseStack[--i]){
		if(clauseStack[i]->assignedValue==Modified){
			resetClauseCount(clauseStack[i]);
			clauseStack[i]->assignedValue=False;
			clauseStack[i]->weight = Weight[clauseStack[i]->currentCount];
			for(j=0;j<clauseStack[i]->bodyCount;j++)
				if((clauseStack[i]->bodyList[j]->assignedValue==Unknown)
					&&(clauseStack[i]->bodyList[j]->nType==ATOM))
						setAtomWeight(clauseStack[i]->bodyList[j]);
			for(j=0;j<clauseStack[i]->headCount;j++)
				if((clauseStack[i]->headList[j]->assignedValue==Unknown)
					&&(clauseStack[i]->headList[j]->nType==ATOM))
						setAtomWeight(clauseStack[i]->headList[j]);
		}
	}
	if(allAtoms > numberOfAtoms)
		metaAtom = getMetaAtom();
	if(metaAtom > 0 )
		setLiteralList(&atomList[metaAtom]); 
	else if((literal=getLiteral())){
		if(atomList[literal].currentHeadCount < atomList[literal].currentBodyCount)
			literal = -literal;
		literalList[1]=literal;
		literalList[0]=-literal;
		pushBranchValues(2);
	}
	else return(0);
	return(1);
}

int DPBacktracking(double begin)
{
	int flag=1,more=1,backtrack = 0;
	struct tms 	cpu;
	double 		cputime=0;

	while(more){	
		if(Limit && flag!=-1){
			times(&cpu);
			cputime=(cpu.tms_utime-begin)/(double)CLK_TCK;
			if(Limit < cputime)
				flag=2;
			Restart=0;
		}
		if(Restart>0 && backtrack > numberOfAtoms && flag!=-1){
			flag = 2;
		}
		switch(flag){
		case 1:
			flag=getBranchLiterals();
			if(Debug){
				fprintf(errorFile, "\ngetBranchLiterls=%i branchTop=%i\n", 
						flag,branchTop);
			}
			if(flag){
				Branch++;
				popBacktrackingStacks();
				flag=setupPropagation();
				if(Debug){
					fprintf(errorFile,
					"\nbranchPropagation=%i  branch=%i Assigned=%i Satisfied=%i requiredAtomTop=%i clauseTop=%i\n ", 
						flag,Branch,atomsAssigned,currentSatisfiedClauses,requiredAtomTop,clauseTop);
				}
			}
			break;
		case 0:
			Backtrack++;
			backtrack++;
			if(popBacktrackingStacks()){
				flag=setupPropagation();
				if(Debug){
					fprintf(errorFile,
						"\nbackTrackpropagation=%i Backtrack=%i  Assigned=%i Satisfied=%i\n ", 
						flag,Backtrack,atomsAssigned,currentSatisfiedClauses);
				}
			}
			else{
				if(Cnt)
					flag=-1;
				more=0;
			}
			break;
		case -1: 
			if(Answer)
				printAnswer();
			setCnt();
			if(Cnt < Count || !Count)
				flag=0;
			else more=0;
			break;
		default :
			more=0;
			break;
		}
	}
	return(flag);
}

int evaluateClause(Node *clause)
{
	int ret=1,horn =0;

	if(clause->assignedValue==True)
		return(1);
	if(clause->nType == HORNCLAUSE)
		horn = 1;
	if(clause->currentBodyCount==0){
		if(clause->currentHeadCount==0){
			ret=0;
		}
		else if(clause->currentHeadCount==1){
			if((ret=propagateAtoms(clause->headList,clause->headCount,horn,1))==1){ 
				pushClauseStack(clause,True);
			}
		}
	}
	else if(clause->currentBodyCount==1&&clause->currentHeadCount==0){
		if((ret=propagateAtoms(clause->bodyList,clause->bodyCount,0,0))==1){
			pushClauseStack(clause,True);
		}
	}
	return(ret);
}



int preprocess(void)
{
	int i;

	memset(propagateQueueValue,-1,sizeof(int)*(allAtoms+1));
	propagateCount=0;
	for(i=0;i<numberOfClauses;i++){
		if(!evaluateClause(&clauseList[i]))
			return(0);
	}
	if(propagateCount)
		if(!propagate())
			return(0);
	memset(propagateQueueValue,-1,sizeof(int)*(allAtoms+1));
	propagateCount=0;
	for(i=1;i<=numberOfAtoms;i++){
		if(atomList[i].nType==HORNATOM&&
				atomList[i].isRequired&&
				atomList[i].requiredValue==True&&
				atomList[i].assignedValue==Unknown&&
				propagateQueueValue[i]==Unknown){
			if(!checkHorn(&atomList[i])) return(0);
		}
	}
	if(propagateCount)
		if(!propagate())
			return(0);
	if(requiredSatisfiedClauses<=currentSatisfiedClauses){
		Cnt++;
		if(Answer)
			printAnswer();
		return(-1);
	}
	return(1);
}
void freeNode(Node *node)
{
	free(node->atomText);
	free(node->metaList);
	free(node->headList);
	free(node->bodyList);
}


void freeLists(void)
{
	int i;

	free(Weight);
	free(propagateQueueValue);
	free(hornPathList);
	free(propagateQueue);
	free(branchingStack);
	free(literalList);
	free(requiredAtomStack);
	free(atomStack);
	free(clauseStack);
	for(i=0;i<=allAtoms;i++)
		freeNode(&atomList[i]);
	for(i=0;i<numberOfClauses;i++)
		freeNode(&clauseList[i]);
	free(clauseList);
	free(atomList);

}


int setUpProblem(char *filename)
{

	FILE 		*summary;
	static char	*solution[]={"unsat","sat","unknown","error"};
	int			answer=0,sol=0;
	struct tms 	cpu;
	struct stat buf;
	double 		cputime=0,begin=0;


	if(getProblem(&sol)==0){
		freeLists();
		return(0);
	}
	if(Print){
		printTheory();
		freeLists();
		return(0);
	}
	times(&cpu);
	begin=cpu.tms_utime;
	answer=0;
	if(sol > 0){
		setWeights();
		setReference();
		do{
			if(Debug)
				fprintf(errorFile,"Restart = %i sol = %i\n ",Restart,sol);
			initializeCurrent();
			if((sol=preprocess())>0)
				sol=DPBacktracking(begin);
			Restart--;
		} while(Restart>=0 && sol >0);
	}
	if(Debug)
		fprintf(errorFile,"sol = %i ",sol);
	times(&cpu);
	cputime=(cpu.tms_utime-begin)/(double)CLK_TCK;
	if(Debug)
		fprintf(errorFile,"cputime = %g ",cputime);
	if(sol==-1)
		sol=1;
	if(stat("aspps.stat",&buf)==-1)
		answer=1;
	if((summary=fopen("aspps.stat","a"))==NULL){
		fprintf(errorFile, "cannot open aspps.stat");
	}
	else{
		if(answer)
			fprintf(summary,"Variable  Clauses   Sat   CPU Time  Branch    Bcktrck  Count     Filename\n");
		fprintf(summary,"%-10i%-10i%-6s%-10g%-10i%-10i%-10i",allAtoms,numberOfClauses,solution[sol],cputime,Branch,Backtrack,Cnt);
		fprintf(summary,"%s\n",filename);
		fclose(summary);
	}
	freeLists();
	return(sol);
}

void useageErrorMessage(void)
{
	fprintf(stdout, "\nuseage: aspps [-f filename] ");
	fprintf(stdout, " [-r] [-L n.n ] [-C [nn]] [-A] [-S predicate] [-R [nn]] \n");
	fprintf(stdout, "-C : finds all solutions.");
	fprintf(stdout, " Number of solutions written to stat file \n");
	fprintf(stdout, "-C nn: finds first nn solutions.\n");
	fprintf(stdout, "-L n.n: Limits execution time to n.n seconds.\n");
	fprintf(stdout, "-A : Displays all positive atoms\n");
	fprintf(stdout, "-r : turns off random choice of equal branching.");
	fprintf(stdout, "-R : Restarts; requires random on.");
	fprintf(stdout, "-S predicate : ");
	fprintf(stdout, " Displays all positive atoms constructed from predicate\n");
}
	

int main(int argc, char **argv){

	char	filename[130];
	int	i,j,fn=0,ret=0;

	Print=0;
	Restart=1;
	currentSatisfiedClauses=0;
	Count=1;
	Cnt=0;
	Answer=0;
	Backtrack=0;
	srand((unsigned)time(NULL));
	getcwd(filename,129); 
	ShowAtoms=NULL;
	printf("\n%s\n",VERSION);
	i=1;
	while(i<argc){
		if(argv[i][0]=='-')
			switch(argv[i][1]){
				case 'P': Print=1;break; 
				case 'r': Random=0;break; 
				case 'f': fn=1;strcpy(filename,argv[++i]);break;
				case 'L': if((i+1==argc)||(argv[i+1][0]=='-')) {
								 useageErrorMessage();
								 return(3);
							 }
						  else Limit=atof(argv[++i]);
						  break;
				case 'C': if((i+1==argc)||(argv[i+1][0]=='-')){
								 if(isdigit(argv[i][1])) 
									  Count= atoi(&argv[i][1]);
							 }
						  else Count=atoi(argv[++i]);
						  break;
				case 'R': if((i+1==argc)||(argv[i+1][0]=='-')){
								 if(isdigit(argv[i][2])) 
									  Restart= atoi(&argv[i][2]);
							 }
						  else Restart=atoi(argv[++i]);
						  break;
				case 'S': while(ShowCnt+i+1<argc&&argv[i+ShowCnt+1][0]!='-') ShowCnt++;
						  if((ShowAtoms=(char **)malloc(sizeof(char *)*ShowCnt))==NULL)
							  exit(1);
						  j=0;
						  while(j<ShowCnt)
							  ShowAtoms[j++]=strdup(argv[++i]);
						  if(ShowCnt)
							  Answer=1;
						  break;
				case 'A': Answer=1;break;
				default: 
					useageErrorMessage();
					return(3);
		}
		i++;
	}
	if(Restart < 1) Restart = 1;
	if(Restart > 1 && Random == 0)
		Random = 1;
	errorFile=stdout;
	if(!fn){
		fptr = stdin;
		ret=setUpProblem("stdin");
	}
	else if((fptr=fopen(filename,"r"))==NULL){
		printf("error: can not open  %s \n",filename);
		return(3);
	}
	else 
		ret=setUpProblem(filename);
	return(ret);
}
